module Main where


----------------------------------------------------------------------
----------------------------------------------------------------------
-- 1. Escreva uma função recursiva elem' que verifica se um valor está
--    contido numa lista

elem' :: Eq a => a -> [a] -> Bool
elem' _ [] = False
elem' e (x:xs)
  | e == x    = True
  | otherwise = elem' e xs

----------------------------------------------------------------------
----------------------------------------------------------------------
-- 2. Defina a função recursiva concat' que recebe uma lista de listas
--    e  devolve uma  única  lista contendo  os  elementos das  listas
--    internas.

-- >>> concat' [[1, 2, 3], [3], [4, 5]]
-- [1,2,3,3,4,5]


-- >>> concat' [[], [1, 2], [3], [4, 5, 6]]
-- [1,2,3,4,5,6]

concat' :: [[a]] -> [a]
concat' []       = []
concat' ([]:xss) =  concat' xss
concat' ((x:xs):xss) = x : concat' (xs:xss)



-- 3. Defina a função que recebe um char separador e uma lista de
--    chars (ou string) e devolve uma string com os elementos
--    intercalados entre o separador. Exemplo:
--
--        > intersperse' '-' "abcde"
--        "a-b-c-d-e"

intersperse' :: a -> [a] -> [a]
intersperse' _ []     = []
intersperse' _ [x]    = [x]
intersperse' s (x:xs) = x : s : intersperse' s xs


----------------------------------------------------------------------
----------------------------------------------------------------------
-- 4. Escreva a função wordNumber que recebe um char separador e um
--    inteiro, e devolve uma string com os nomes dos números separados
--    por um char sep. Você poderá utilizar as funções concat' e
--    intersperse' definidas anteriormente. Outras funções auxiliares
--    podem ser definidas, como transformar um inteiro em uma lista de
--    dígitos, e substituir cada dígito pelo seu correspondente por
--    extenso. Exemplo:
--
--        > wordNumber '-' 123
--        "um-dois-três"

numToList :: Int -> [Int]
numToList n
  | n == 0 = []
  | otherwise = numToList (n `div` 10) ++ [n `mod` 10]

digitName :: Int -> String
digitName n  = nomes !! n
  where
    nomes = ["zero", "um", "dois", "tres", "quatro", "cinco", "seis", "sete" , "oito", "nove"]

wordNumber :: Char -> Int -> String
wordNumber s n =
  concat' inter
  where
  nomes  = map digitName (numToList n)
  inter = intersperse' [s] nomes

-- >>>  wordNumber '-' 123
-- "um-dois-tres"

----------------------------------------------------------------------
----------------------------------------------------------------------
-- 5 Defina uma função recursiva merge
--    que recebe duas listas ordenadas e devolve uma única lista ordenada.
--    Exemplo:
--        > merge [2,5,6] [1,3,4]
--          [1, 2, 3, 4, 5, 6]

merge :: Ord a => [a] -> [a] -> [a]
merge [] [] = []
merge xs [] = xs
merge [] ys = ys
merge (x:xs) (y:ys)
  | x <= y    = x : merge xs (y:ys)
  | otherwise = y : merge (x:xs) ys

-- >>> merge [2,5,6] [1,3,4]
-- [1,2,3,4,5,6]

----------------------------------------------------------------------
----------------------------------------------------------------------
-- 6.Utilizando a função merge implementada anteriormente, defina

msort :: Ord a => [a] -> [a]
msort []  = []
msort [x] = [x]
msort xs =
  merge (msort left) (msort right)
  where
    (left, right) = halve xs

-- que implementa o algoritmo merge sort. Para auxiliar, defina halve
-- que divide uma lista em duas de mesmo tamanho ou com uma diferença
-- máxima de um elemento.

halve :: [a] -> ([a],[a])
halve []     = ([],[])
halve (x:[]) = ([x],[])
halve xs = splitAt (length xs `div` 2) xs


----------------------------------------------------------------------
----------------------------------------------------------------------
-- 7. Sem olhar as definições do Prelude padrão, defina as seguintes
--    funções de alta-ordem:

-- Verifica se todos elementos da lista satisfazem um predicado
all' :: (a -> Bool) -> [a] -> Bool
all' p = and . map p

-- Verifica se pelo menos um dos elementos da lista satisfazem um predicado
any' :: (a -> Bool) -> [a] -> Bool
any' p xs = or $ map p xs

-- Seleciona os elementos da lista enquanto eles satisfizerem um predicado
takeWhile' :: (a -> Bool) -> [a] -> [a]
takeWhile' _ []     = []
takeWhile' p (x:xs)
  | p x       = x : takeWhile' p xs
  | otherwise = []

-- >>> takeWhile' (> 10) [0..100]
-- []

-- Remove os elementos da lista enquanto eles satisfizerem um predicado
dropWhile' :: (a -> Bool) -> [a] -> [a]
dropWhile' _ []     = []
dropWhile' p (x:xs)
  | p x       = dropWhile' p xs
  | otherwise = x:xs

-- >>> dropWhile' (< 50) [0..100]
-- [50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,97,98,99,100]

----------------------------------------------------------------------
----------------------------------------------------------------------
-- 8. Defina uma função altMap que aplica alternadamente as duas
-- funções recebidas como argumento em uma lista. Exemplo:
--
-- > altMap (+10) (+100) [0,1,2,3,4]
-- [10,101,12,103,14]

altMap :: (a -> b) -> (a -> b) -> [a] -> [b]
altMap _ _ [] = []
altMap f g (x:xs) = f x : altMap g f xs

-- >>> altMap (+10) (+100) [0,1,2,3,4]
-- [10,101,12,103,14]


----------------------------------------------------------------------
----------------------------------------------------------------------
-- 9. Usando /folding/, defina a função

foldl2 :: (a -> b -> a) -> a -> [b] -> a
foldl2 _ v []     = v
foldl2 f v (x:xs) = foldl2 f (f v x) xs

foldr2 :: (b -> a -> a) -> a -> [b] -> a
foldr2 _ v []     = v
foldr2 f v (x:xs) = f x (foldr2 f v xs)


dec2int :: [Int] -> Int
dec2int = foldl (\x y -> 10 * x + y) 0

-- que converte uma lista de inteiros em um inteiro. Exemplo:
--     > dec2int [2,3,4,5]
--     2345

----------------------------------------------------------------------
----------------------------------------------------------------------
-- 10. Utilize /folding/ para definir as seguintes funções em Haskell:
-- retorna True se pelo menos um booleano da lista for True

or' :: [Bool] -> Bool
or' = foldl (||) False

-- inverte os elementos de uma lista
reverse' :: [a] -> [a]
reverse' = foldl (flip (:)) []

-- >>> reverse' [0..10]
-- [10,9,8,7,6,5,4,3,2,1,0]

-- filtra os elementos de uma lista de acordo com um predicado
filter' :: (a -> Bool) -> [a] -> [a]
filter' p = foldr (\x acc -> if p x then x : acc else acc) []

-- >>> filter' even [0..10]
-- [0,2,4,6,8,10]




----------------------------------------------------------------------
----------------------------------------------------------------------
-- 11. Escreva duas versões para a função elem''
-- que verifica se um elemento está em uma lista. Uma versão deverá
-- utilizar /folding/ e outra a função /any/.

elem'' :: Eq a => a -> [a] -> Bool
elem'' v xs = foldl (\res x -> if x == v then True else res) False xs

elem''' :: Eq a => a -> [a] -> Bool
elem''' x xs = any (==x) xs

elem'''' :: Eq a => a -> [a] -> Bool
elem'''' = any . (==)

main :: IO ()
main = do
  putStrLn "hello world"
